#ifndef MENU_LIB_H
#define MENU_LIB_H

#include "object.h"

#include <string>
#include <vector>

namespace pwn {

//extension of object that resizes itself around the text provided
class Text : public Object {
	typedef Object super;
public:
	Text(const vec2f &_pos = vec2f(0,0), const char *_str = "");
};


class Scrollbar;
class ScrollbarListener {
public:
	virtual void callback_onSliderChange(Scrollbar *sb) = 0;
};

class Scrollbar : public Object {
	typedef Object super;
protected:
	std::vector<ScrollbarListener* > scrollbarListeners;
	float sliderPos;
	
	gc::Pointer<Text> scrollUL;
	gc::Pointer<Text> scrollDR;
	gc::Pointer<Object> scrollTab;
	bool horz;
	float len;
	float min;
	float max;
	float speed;
public:

	Scrollbar(const vec2f &_pos, float len, bool horz, float min, float max, float sliderPos);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);

	//interface for adding listeners
	virtual void addScrollbarListener(ScrollbarListener *l) {
		scrollbarListeners.push_back(l);
	}

	//somethign i've experimented with in my mind:
	//instead, make 'sliderPos' an object for which the = operator is redefined
	//of course the problem here is yo'ud have to also redefine all the other operators in order to recreate full variable operator control
	virtual float getSliderPos() const { return sliderPos; }
	virtual void setSliderPos(float pos, bool forceReset = false);
	virtual vec2f getRange() const { return vec2f(min,max); }
	virtual void setRange(float min, float max);
	void setRange(const vec2f &r) { setRange(r(0), r(1)); }
	
	virtual void setSize(const vec2f &_size);
	virtual void setLength(float _len);

	virtual bool isHorz() const { return horz; }
	virtual float getLength() const { return len; }
	virtual float getMin() const { return min; }
	virtual float getMax() const { return max; }

	virtual Object *getScrollTab() { return scrollTab; }
};


class CloseButton : public Text {
typedef Text super;
protected:
	bool hideInstead;
public:
	CloseButton(const vec2f &_pos, bool _hideInstead);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
	virtual void closeParent();	//always RESPONSE_STOP from whatever you're doing when you call this
};


class Window : public Object {
	typedef Object super;
public:
	Window(const vec2f &_pos, const vec2f &_size, bool close = true);

	bool allowHide;			//allow right-click-to-hide.  overrides allowClose atm
	bool allowClose;			//allow right-click to delete
	bool allowMove;			//allow left-click to move
	gc::Pointer<CloseButton> closeButton;		//if a close button exists

	virtual void setSize(const vec2f &_size);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

class Checkbox;
class CheckboxListener {
public:
	virtual void callback_onSetChecked(Checkbox *m, bool isChecked) = 0;
};
class Checkbox : public Text {
	typedef Text super;
protected:
	std::vector<CheckboxListener*> checkboxListeners;
	bool checked;
public:
	Checkbox(const vec2f &_pos, const char *_str, bool _checked = false);

	//interface for adding listeners
	virtual void addCheckboxListener(CheckboxListener *l) {
		checkboxListeners.push_back(l);
	}

	virtual bool display();
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);

	virtual void setChecked(bool ch);
	virtual bool isChecked() { return checked; }
};

class RadioGroup;
class Radio : public Checkbox {
	typedef Checkbox super;
public:
	//maybe i should stop using 'checked'
	//and group should hold the currently-selected
	gc::Pointer<RadioGroup> group;

	Radio(const vec2f &_pos, const char *_str, RadioGroup *_group, bool _checked = false);
	virtual void setChecked(bool ch);

	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

class RadioGroup : public gc::Object {
public:
	std::vector<gc::Pointer<Radio> > members;
	
	//should only be called by Radio objects
	void clearAllExcept(Radio *r);
};


/*
editable text field input
*/
class TextField;
//listener class for reporting callbacks when the user sets the selected
class TextListener {
public:
	virtual void callback_onTextChange(TextField *) = 0;
};
class TextField : public Object {
	typedef Object super;
protected:
	std::vector<TextListener *> textListeners;
public:
	TextField(const vec2f &_pos, float _width) : super(_pos, vec2f(_width,1)), textHasChanged(false) {
		allowFocus = true;
	}
	
	//interface for adding listeners
	virtual void addTextListener(TextListener *l) {
		textListeners.push_back(l);
	}

	//this functionality is all in common with Select for allowEdit == true
	
	//used with allowEdit
	//set to false on focus
	//set true again when the user presses a key and changes something
	//then, when we lose focus, if this is set, then call a 'got a text change' function
	//you know, this has a lot in common with Select - when it is editable, at least
	bool textHasChanged;
	
	virtual void onFocus();
	virtual void onBlur();
	virtual bool doKeyPress(int key); //do something cool here.
	virtual void onTextChange() {}
};

//name is already used...
class Select;

/*
containing object within a selection box
*/
//you know, if not for the MenuEditorWindow, this would be so easy to externalize . . .
//hmm, how can we make menu_select_contents a window but still allow editor specific ones?
//1) make a general non-editor menu_select_contents
//2) for the editor make a special subclass
class SelectContents : public Window {	//notice - this used to be MenuEditorWindow
	typedef Window super;
public:
	gc::Pointer<Select> owningSelect;	//here's another apple gcc folly
	
	SelectContents(const vec2f &_pos, const vec2f &_size, Select *_owningSelect);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

//listener class for reporting callbacks when the user sets the selected
class SelectListener {
public:
	virtual void callback_userSetSelected(Select *, int newSel, int oldSel) = 0;
};

/*
a selection box.  something resembling a combo box at the moment . . .
*/
class Select : public TextField {
	typedef TextField super;
protected:
	//keep a list of everyone attached to us as a SelectListener
	std::vector<SelectListener *> selectListeners;
	
public:
	//keep track of all entries *HERE*
	//use this for refreshing ,contents, and for refreshing our own text
	std::vector<gc::Pointer<Text> > items;
	
	//interface for adding listeners
	virtual void addSelectListener(SelectListener *l) {
		selectListeners.push_back(l);
	}

	//which one is showing at present:
	int selectedIndex;
	
	//whether or not we should allow fallback on our TextField parent
	//maybe this should be a part of TextField as a means of enable/disable
	bool allowEdit;
	
	//used with allowEdit
	//set to false on focus
	//set true again when the user presses a key and changes something
	//then, when we lose focus, if this is set, then call a 'got a text change' function
	bool textHasChanged;
	
	//holding all our items
	gc::Pointer<SelectContents> contents;
	
	//starts off with no entries
	Select(const vec2f &_pos, float width);
	
	virtual void setSelected(int index);		//setting the selected:
	virtual void resetContent();			//reset all entries
	virtual void addString(const char *str);	//on add, ammend to contents
	const char *getString(int index);
	
	//on click, reposition contents and show it
	//on contents click, do some callback
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);

	virtual void onFocus();
	virtual void onBlur();

	//if the user sets the selection . . .
	virtual void userSetSelected(int _sel);

	virtual bool doKeyPress(int key); //if allowEdit is true then this does something
	
	//if the user types something into the box . . .
	virtual void onTextChange() {}
	
};

//also overriding for callback's sake - upon setsize
class ScrollArea;
class ScrollContainer : public Object {
	typedef Object super;
public:
	ScrollContainer(const vec2f &_pos = vec2f(0,0), const vec2f &_size = vec2f(1,1));
	
	//dont let this set 'size', but instead fall back to 'truesize'
	virtual void setSize(const vec2f &_size);

	//attaching this will set this variable
	gc::Pointer<ScrollArea> scrollArea;
};

class TextScrollContainer : public ScrollContainer {
	typedef ScrollContainer super;
public:
	TextScrollContainer(const vec2f &_pos, const vec2f &_size);
	
	virtual void setText(const std::string &_str);
};

//overriding for callbacks'sake . . right now this is the current class of horz & vert below . . .
class ScrollAreaBar : public Scrollbar {
	typedef Scrollbar super;
public:
	ScrollAreaBar(const vec2f &_pos, float len, bool horz, float min, float max, float sliderPos, ScrollArea *_parent);
	virtual void setSliderPos(float pos, bool forceReset = false);
	gc::Pointer<ScrollArea> parent;
};

/*
scroll pane for which the child size and the visible size are dynamically controllable
*/
class ScrollArea : public Object {
	typedef Object super;
protected:
	gc::Pointer<Scrollbar> horz, vert;
	gc::Pointer<ScrollContainer> scrollPanel;
public:
	ScrollArea(const vec2f &_pos = vec2f(0,0), const vec2f &_size = vec2f(1,1), ScrollContainer *_scrollPanel = NULL);

	virtual void updateChildSize();
	virtual void setSize(const vec2f &_size);
	
	ScrollContainer *getScrollPanel() { return scrollPanel.ptr(); }
	Scrollbar *getHorz() { return horz.ptr(); }
	Scrollbar *getVert() { return vert.ptr(); }
};

//this is just a mess so i'm not even going to touch it at the moment
//SubMenu and BaseMenuEntry need to be merged (they are 1-1 anyways)
//and then it'll parallel java
//but we can (always) automate further by just making one class tree
//and have it step right on the first level and down for every subsequent onew

class SubMenuEntry : public Text {
	typedef Text super;
public:
	int ident;
	void (*callback)(int);	//callback to call with 'ident' in hand.  parallels Win32 pretty closely
	SubMenuEntry(const char *_string, int _ident = 0, void (*_callback)(int) = NULL, const vec2f &_pos = vec2f(0,0));
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

class BaseMenuEntry;
//this holds the entries that span down
class SubMenu : public Window {
	typedef Window super;
protected:
	std::list<gc::Pointer<SubMenuEntry> > subs;
	friend class BaseMenuEntry;
	BaseMenuEntry *baseEntry;
public:
	SubMenu();
	SubMenu(void (*entry_callback)(int), ...);
	virtual void addChild(SubMenuEntry *);
protected:
	void setBaseMenu(BaseMenuEntry *);
};

class BaseMenuEntry : public Text {
	typedef Text super;
public:
	gc::Pointer<SubMenu> submenu;
	
	//odd setup
	//rootmenu is the menu we attach this to
	//it cant go on the BaseMenu, for clipping reasons
	//maybe i need to toss away the whole clipping system?  or think up a new one?
	BaseMenuEntry(const char *_string, SubMenu *_submenu, Object * rootmenu);
};

//this holds the entries that span to the right
class BaseMenu : public Window {
	typedef Window super;
	std::list<gc::Pointer<BaseMenuEntry> > subs;
public:
	gc::Pointer<SubMenu> lastsel;		//the last opened submenu (child of the base entry)

	BaseMenu(const vec2f &_pos = vec2f());
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
	virtual void hideLast();
	virtual void addChild(BaseMenuEntry *);
};

};

#endif	//MENU_LIB_H
